import dataclasses
import typing
from contextlib import contextmanager

import fntypes.option
import fntypes.result
import msgspec
from fntypes.co import Error, Ok, Result, Variative

if typing.TYPE_CHECKING:
    from datetime import datetime

    from fntypes.option import Option

    def get_class_annotations(obj: typing.Any) -> dict[str, type[typing.Any]]: ...

    def get_type_hints(obj: typing.Any) -> dict[str, type[typing.Any]]: ...

else:
    from datetime import datetime as dt

    from msgspec._utils import get_class_annotations, get_type_hints

    Value = typing.TypeVar("Value")
    Err = typing.TypeVar("Err")

    datetime = type("datetime", (dt,), {})

    class OptionMeta(type):
        def __instancecheck__(cls, __instance: typing.Any) -> bool:
            return isinstance(__instance, fntypes.option.Some | fntypes.option.Nothing)

    class Option(typing.Generic[Value], metaclass=OptionMeta):
        pass


T = typing.TypeVar("T")

DecHook: typing.TypeAlias = typing.Callable[[type[T], typing.Any], typing.Any]
EncHook: typing.TypeAlias = typing.Callable[[T], typing.Any]

Nothing: typing.Final[fntypes.option.Nothing] = fntypes.option.Nothing()


def get_origin(t: type[T]) -> type[T]:
    return typing.cast(T, typing.get_origin(t)) or t


def repr_type(t: typing.Any) -> str:
    return getattr(t, "__name__", repr(get_origin(t)))


def is_common_type(type_: typing.Any) -> typing.TypeGuard[type[typing.Any]]:
    if not isinstance(type_, type):
        return False
    return (
        type_ in (str, int, float, bool, None, Variative)
        or issubclass(type_, msgspec.Struct)
        or hasattr(type_, "__dataclass_fields__")
    )


def type_check(obj: typing.Any, t: typing.Any) -> bool:
    return (
        isinstance(obj, t)
        if isinstance(t, type) and issubclass(t, msgspec.Struct)
        else type(obj) in t
        if isinstance(t, tuple)
        else type(obj) is t
    )


def msgspec_convert(obj: typing.Any, t: type[T]) -> Result[T, str]:
    try:
        return Ok(decoder.convert(obj, type=t, strict=True))
    except msgspec.ValidationError:
        return Error(
            "Expected object of type `{}`, got `{}`.".format(
                repr_type(t),
                repr_type(type(obj)),
            )
        )


def msgspec_to_builtins(
    obj: typing.Any,
    *,
    str_keys: bool = False,
    builtin_types: typing.Iterable[type[typing.Any]] | None = None,
    order: typing.Literal["deterministic", "sorted"] | None = None,
) -> fntypes.result.Result[typing.Any, msgspec.ValidationError]:
    try:
        return Ok(encoder.to_builtins(**locals()))
    except msgspec.ValidationError as exc:
        return Error(exc)


def option_dec_hook(tp: type[Option[typing.Any]], obj: typing.Any) -> Option[typing.Any]:
    if obj is None:
        return Nothing

    (value_type,) = typing.get_args(tp) or (typing.Any,)
    orig_value_type = typing.get_origin(value_type) or value_type
    orig_obj = obj

    if not isinstance(orig_obj, dict | list) and is_common_type(orig_value_type):
        if orig_value_type is Variative:
            obj = value_type(orig_obj)  # type: ignore
            orig_value_type = typing.get_args(value_type)

        if not type_check(orig_obj, orig_value_type):
            raise TypeError(f"Expected `{repr_type(orig_value_type)}`, got `{repr_type(type(orig_obj))}`.")

        return fntypes.option.Some(obj)

    return fntypes.option.Some(decoder.convert(orig_obj, type=value_type))


def variative_dec_hook(tp: type[Variative], obj: typing.Any) -> Variative:
    union_types = typing.get_args(tp)

    if isinstance(obj, dict):
        models_struct_fields: dict[type[msgspec.Struct], int] = {
            m: sum(1 for k in obj if k in m.__struct_fields__)
            for m in union_types
            if issubclass(get_origin(m), msgspec.Struct)
        }
        union_types = tuple(t for t in union_types if t not in models_struct_fields)
        reverse = False

        if len(set(models_struct_fields.values())) != len(models_struct_fields.values()):
            models_struct_fields = {m: len(m.__struct_fields__) for m in models_struct_fields}
            reverse = True

        union_types = (
            *sorted(
                models_struct_fields,
                key=lambda k: models_struct_fields[k],
                reverse=reverse,
            ),
            *union_types,
        )

    for t in union_types:
        if not isinstance(obj, dict | list) and is_common_type(t) and type_check(obj, t):
            return tp(obj)
        match msgspec_convert(obj, t):
            case Ok(value):
                return tp(value)

    raise TypeError(
        "Object of type `{}` does not belong to types `{}`".format(
            repr_type(obj.__class__),
            " | ".join(map(repr_type, union_types)),
        )
    )


@typing.runtime_checkable
class DataclassInstance(typing.Protocol):
    __dataclass_fields__: typing.ClassVar[dict[str, dataclasses.Field[typing.Any]]]


class Decoder:
    """Class `Decoder` for `msgspec` module with decode hook
    for objects with the specified type.

    ```
    import enum

    from datetime import datetime as dt

    class Digit(enum.IntEnum):
        ONE = 1
        TWO = 2
        THREE = 3

    decoder = Encoder()
    decoder.dec_hooks[dt] = lambda t, timestamp: t.fromtimestamp(timestamp)

    decoder.dec_hook(dt, 1713354732)  #> datetime.datetime(2024, 4, 17, 14, 52, 12)

    decoder.convert("123", type=int, strict=False)  #> 123
    decoder.convert(1, type=Digit)  #> <Digit.ONE: 1>

    decoder.decode(b'{"digit":3}', type=dict[str, Digit])  #> {'digit': <Digit.THREE: 3>}
    ```
    """

    def __init__(self) -> None:
        self.dec_hooks: dict[typing.Any, DecHook[typing.Any]] = {
            Option: option_dec_hook,
            Variative: variative_dec_hook,
            datetime: lambda t, obj: t.fromtimestamp(obj),
            fntypes.option.Some: option_dec_hook,
            fntypes.option.Nothing: option_dec_hook,
        }

    def __repr__(self) -> str:
        return "<{}: dec_hooks={!r}>".format(
            self.__class__.__name__,
            self.dec_hooks,
        )

    @typing.overload
    def __call__(self, type: type[T]) -> typing.ContextManager[msgspec.json.Decoder[T]]: ...

    @typing.overload
    def __call__(self, type: typing.Any) -> typing.ContextManager[msgspec.json.Decoder[typing.Any]]: ...

    @typing.overload
    def __call__(
        self, type: type[T], *, strict: bool = True
    ) -> typing.ContextManager[msgspec.json.Decoder[T]]: ...

    @typing.overload
    def __call__(
        self, type: typing.Any, *, strict: bool = True
    ) -> typing.ContextManager[msgspec.json.Decoder[typing.Any]]: ...

    @contextmanager
    def __call__(self, type=object, *, strict=True):
        """Context manager returns the `msgspec.json.Decoder` object with the `dec_hook`."""

        dec_obj = msgspec.json.Decoder(
            type=typing.Any if type is object else type, strict=strict, dec_hook=self.dec_hook
        )
        yield dec_obj

    def add_dec_hook(self, t: T):  # type: ignore
        def decorator(func: DecHook[T]) -> DecHook[T]:
            return self.dec_hooks.setdefault(get_origin(t), func)  # type: ignore

        return decorator

    def dec_hook(self, tp: type[typing.Any], obj: object) -> object:
        origin_type = t if isinstance((t := get_origin(tp)), type) else type(t)
        if origin_type not in self.dec_hooks:
            raise TypeError(
                f"Unknown type `{repr_type(origin_type)}`. You can implement decode hook for this type."
            )
        return self.dec_hooks[origin_type](tp, obj)

    def convert(
        self,
        obj: object,
        *,
        type: type[T] = dict,
        strict: bool = True,
        from_attributes: bool = False,
        builtin_types: typing.Iterable[type[typing.Any]] | None = None,
        str_keys: bool = False,
    ) -> T:
        return msgspec.convert(
            obj,
            type,
            strict=strict,
            from_attributes=from_attributes,
            dec_hook=self.dec_hook,
            builtin_types=builtin_types,
            str_keys=str_keys,
        )

    @typing.overload
    def decode(self, buf: str | bytes) -> typing.Any: ...

    @typing.overload
    def decode(self, buf: str | bytes, *, type: type[T]) -> T: ...

    @typing.overload
    def decode(self, buf: str | bytes, *, type: typing.Any) -> typing.Any: ...

    @typing.overload
    def decode(
        self,
        buf: str | bytes,
        *,
        type: type[T],
        strict: bool = True,
    ) -> T: ...

    @typing.overload
    def decode(
        self,
        buf: str | bytes,
        *,
        type: typing.Any,
        strict: bool = True,
    ) -> typing.Any: ...

    def decode(self, buf, *, type=object, strict=True):
        return msgspec.json.decode(
            buf,
            type=typing.Any if type is object else type,
            strict=strict,
            dec_hook=self.dec_hook,
        )


class Encoder:
    """Class `Encoder` for `msgspec` module with encode hooks for objects.

    ```
    from datetime import datetime as dt

    encoder = Encoder()
    encoder.enc_hooks[dt] = lambda d: int(d.timestamp())

    encoder.enc_hook(dt.now())  #> 1713354732
    encoder.encode({'digit': Digit.ONE})  #> '{"digit":1}'
    ```
    """

    def __init__(self) -> None:
        self.enc_hooks: dict[typing.Any, EncHook[typing.Any]] = {
            fntypes.option.Some: lambda opt: opt.value,
            fntypes.option.Nothing: lambda _: None,
            Variative: lambda variative: variative.v,
            datetime: lambda date: int(date.timestamp()),
        }

    def __repr__(self) -> str:
        return "<{}: enc_hooks={!r}>".format(
            self.__class__.__name__,
            self.enc_hooks,
        )

    @contextmanager
    def __call__(
        self,
        *,
        decimal_format: typing.Literal["string", "number"] = "string",
        uuid_format: typing.Literal["canonical", "hex"] = "canonical",
        order: typing.Literal[None, "deterministic", "sorted"] = None,
    ) -> typing.Generator[msgspec.json.Encoder, typing.Any, None]:
        """Context manager returns the `msgspec.json.Encoder` object with the `enc_hook`."""

        enc_obj = msgspec.json.Encoder(enc_hook=self.enc_hook)
        yield enc_obj

    def add_dec_hook(self, t: type[T]):
        def decorator(func: EncHook[T]) -> EncHook[T]:
            encode_hook = self.enc_hooks.setdefault(get_origin(t), func)
            return func if encode_hook is not func else encode_hook

        return decorator

    def enc_hook(self, obj: object) -> object:
        origin_type = get_origin(obj.__class__)
        if origin_type not in self.enc_hooks:
            raise NotImplementedError(
                f"Not implemented encode hook for object of type `{repr_type(origin_type)}`."
            )
        return self.enc_hooks[origin_type](obj)

    @typing.overload
    def encode(self, obj: typing.Any) -> str: ...

    @typing.overload
    def encode(self, obj: typing.Any, *, as_str: typing.Literal[True]) -> str: ...

    @typing.overload
    def encode(self, obj: typing.Any, *, as_str: typing.Literal[False]) -> bytes: ...

    def encode(self, obj: typing.Any, *, as_str: bool = True) -> str | bytes:
        buf = msgspec.json.encode(obj, enc_hook=self.enc_hook)
        return buf.decode() if as_str else buf

    def to_builtins(
        self,
        obj: typing.Any,
        *,
        str_keys: bool = False,
        builtin_types: typing.Iterable[type[typing.Any]] | None = None,
        order: typing.Literal["deterministic", "sorted"] | None = None,
    ) -> typing.Any:
        return msgspec.to_builtins(
            obj,
            str_keys=str_keys,
            builtin_types=builtin_types,
            enc_hook=self.enc_hook,
            order=order,
        )


decoder: typing.Final[Decoder] = Decoder()
encoder: typing.Final[Encoder] = Encoder()


__all__ = (
    "Decoder",
    "Encoder",
    "Nothing",
    "Option",
    "datetime",
    "decoder",
    "encoder",
    "get_class_annotations",
    "get_type_hints",
    "msgspec_convert",
    "msgspec_to_builtins",
)



import hashlib
import hmac
import typing


def verify_webapp_request(
    secret_token: str,
    request_headers: typing.Mapping[str, typing.Any],
) -> bool:
    """Verifies update request is from telegram."""

    return request_headers.get("X-Telegram-Bot-Api-Secret-Token") == secret_token


def webapp_validate_request(
    bot_token: str,
    request_query_params: typing.Mapping[str, typing.Any],
) -> bool:
    """Verifies authentity of webapp request by counting hash of its parameters."""

    items = sorted(request_query_params.items(), key=lambda kv: kv[0])
    data_check_string = "\n".join(f"{k}={param}" for k, param in items if k != "hash")
    secret = hmac.new(
        "WebAppData".encode(),
        bot_token.encode(),
        hashlib.sha256,
    ).digest()
    data_chk = hmac.new(secret, data_check_string.encode(), hashlib.sha256)
    return data_chk.hexdigest() == request_query_params.get("hash")


__all__ = ("verify_webapp_request", "webapp_validate_request")



import asyncio
import dataclasses
import datetime
import typing

from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.dispatch.handler.abc import ABCHandler
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.model import Model

if typing.TYPE_CHECKING:
    from .actions import WaiterActions


T = typing.TypeVar("T", bound=Model)
EventModel = typing.TypeVar("EventModel", bound=BaseCute)

Behaviour: typing.TypeAlias = ABCHandler[T] | None


class ShortStateContext(typing.Generic[EventModel], typing.NamedTuple):
    event: EventModel
    context: Context


@dataclasses.dataclass(slots=True)
class ShortState(typing.Generic[EventModel]):
    event: asyncio.Event
    actions: "WaiterActions"

    release: ABCRule | None = None
    filter: ABCRule | None = None
    lifetime: dataclasses.InitVar[datetime.timedelta | None] = dataclasses.field(
        default=None,
        kw_only=True,
    )

    expiration_date: datetime.datetime | None = dataclasses.field(init=False, kw_only=True)
    creation_date: datetime.datetime = dataclasses.field(init=False)

    context: ShortStateContext[EventModel] | None = dataclasses.field(default=None, init=False, kw_only=True)

    def __post_init__(self, expiration: datetime.timedelta | None = None) -> None:
        self.creation_date = datetime.datetime.now()
        self.expiration_date = (self.creation_date + expiration) if expiration is not None else None

    def cancel(self) -> None:
        """Cancel schedule waiters."""

        waiters = typing.cast(
            typing.Iterable[asyncio.Future[typing.Any]],
            self.event._waiters,  # type: ignore
        )
        for future in waiters:
            future.cancel()


__all__ = ("ShortState", "ShortStateContext")



import asyncio
import typing

import aiohttp
import msgspec
from fntypes.result import Error, Ok

from telegrinder.api.api import API
from telegrinder.api.error import InvalidTokenError
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.modules import logger
from telegrinder.msgspec_utils import decoder
from telegrinder.types.objects import Update, UpdateType


class Polling(ABCPolling):
    def __init__(
        self,
        api: API,
        *,
        offset: int = 0,
        reconnection_timeout: float = 5,
        max_reconnetions: int = 10,
        include_updates: set[str | UpdateType] | None = None,
        exclude_updates: set[str | UpdateType] | None = None,
    ) -> None:
        self.api = api
        self.allowed_updates = self.get_allowed_updates(
            include_updates=include_updates,
            exclude_updates=exclude_updates,
        )
        self.reconnection_timeout = 5 if reconnection_timeout < 0 else reconnection_timeout
        self.max_reconnetions = 10 if max_reconnetions < 0 else max_reconnetions
        self.offset = offset
        self._stop = False

    def __repr__(self) -> str:
        return (
            "<{}: with api={!r}, stopped={}, offset={}, allowed_updates={!r}, "
            "max_reconnetions={}, reconnection_timeout={}>"
        ).format(
            self.__class__.__name__,
            self.api,
            self._stop,
            self.offset,
            self.allowed_updates,
            self.max_reconnetions,
            self.reconnection_timeout,
        )

    @staticmethod
    def get_allowed_updates(
        *,
        include_updates: set[str | UpdateType] | None = None,
        exclude_updates: set[str | UpdateType] | None = None,
    ) -> list[str]:
        allowed_updates: list[str] = list(x.value for x in UpdateType)
        if not include_updates and not exclude_updates:
            return allowed_updates

        if include_updates and exclude_updates:
            allowed_updates = [
                x for x in allowed_updates if x in include_updates and x not in exclude_updates
            ]
        elif exclude_updates:
            allowed_updates = [x for x in allowed_updates if x not in exclude_updates]
        elif include_updates:
            allowed_updates = [x for x in allowed_updates if x in include_updates]

        return [x.value if isinstance(x, UpdateType) else x for x in allowed_updates]

    async def get_updates(self) -> msgspec.Raw | None:
        raw_updates = await self.api.request_raw(
            "getUpdates",
            {
                "offset": self.offset,
                "allowed_updates": self.allowed_updates,
            },
        )
        match raw_updates:
            case Ok(value):
                return value
            case Error(err) if err.code in (401, 404):
                raise InvalidTokenError("Token seems to be invalid")

    async def listen(self) -> typing.AsyncGenerator[list[Update], None]:
        logger.debug("Listening polling")
        reconn_counter = 0

        while not self._stop:
            try:
                updates = await self.get_updates()
                reconn_counter = 0
                if not updates:
                    continue
                updates_list: list[Update] = decoder.decode(updates, type=list[Update])
                if updates_list:
                    yield updates_list
                    self.offset = updates_list[-1].update_id + 1
            except InvalidTokenError as e:
                logger.error(e)
                self.stop()
                exit(3)
            except asyncio.CancelledError:
                logger.info("Caught cancel, polling stopping...")
                self.stop()
            except (aiohttp.client.ServerConnectionError, TimeoutError):
                if reconn_counter > self.max_reconnetions:
                    logger.error(
                        "Failed to reconnect to the server after {} attempts, polling stopping.",
                        self.max_reconnetions,
                    )
                    self.stop()
                    exit(6)
                else:
                    logger.warning(
                        "Server disconnected, waiting 5 seconds to reconnect...",
                    )
                    reconn_counter += 1
                    await asyncio.sleep(self.reconnection_timeout)
            except (aiohttp.ClientConnectorError, aiohttp.ClientOSError):
                logger.error("Client connection failed, attempted to reconnect...")
                await asyncio.sleep(self.reconnection_timeout)
            except BaseException as e:
                logger.exception(e)

    def stop(self) -> None:
        self._stop = True


__all__ = ("Polling",)



import typing

from fntypes.result import Error, Ok, Result

from telegrinder.api import API
from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter.abc import ABCAdapter
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.msgspec_utils import Nothing
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Model, Update

ToCute = typing.TypeVar("ToCute", bound=BaseCute)


class EventAdapter(ABCAdapter[Update, ToCute]):
    ADAPTED_VALUE_KEY: str = "_adapted_cute_event"

    def __init__(self, event: UpdateType | type[Model], cute_model: type[ToCute]) -> None:
        self.event = event
        self.cute_model = cute_model

    def __repr__(self) -> str:
        if isinstance(self.event, str):
            raw_update_type = Update.__annotations__.get(self.event, "Unknown")
            raw_update_type = (
                typing.get_args(raw_update_type)[0].__forward_arg__
                if typing.get_args(raw_update_type)
                else raw_update_type
            )
        else:
            raw_update_type = self.event.__name__

        return "<{}: adapt Update -> {} -> {}>".format(
            self.__class__.__name__,
            raw_update_type,
            self.cute_model.__name__,
        )

    async def adapt(self, api: API, update: Update, context: Context) -> Result[ToCute, AdapterError]:
        if self.ADAPTED_VALUE_KEY in context:
            return Ok(context[self.ADAPTED_VALUE_KEY])

        if isinstance(self.event, UpdateType):
            if update.update_type != self.event:
                return Error(
                    AdapterError(f"Update is not of event type {self.event!r}."),
                )

            if isinstance(event := getattr(update, self.event.value, Nothing), type(Nothing)):
                return Error(
                    AdapterError(f"Update is not an {self.event!r}."),
                )

            event = event.unwrap()

            if type(event) is self.cute_model:
                adapted = event
            else:
                adapted = self.cute_model.from_update(event, bound_api=api)
        else:
            event = getattr(update, update.update_type.value).unwrap()
            if not update.update_type or not issubclass(type(event), self.event):
                return Error(AdapterError(f"Update is not an {self.event.__name__!r}."))

            if type(event) is self.cute_model:
                adapted = event
            else:
                adapted = self.cute_model.from_update(event, bound_api=api)

        context[self.ADAPTED_VALUE_KEY] = adapted
        return Ok(adapted)  # type: ignore


__all__ = ("EventAdapter",)



import typing_extensions as typing
from fntypes.result import Error, Ok, Result

from telegrinder.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.bot.rules.adapter.abc import ABCAdapter, Event
from telegrinder.bot.rules.adapter.errors import AdapterError
from telegrinder.msgspec_utils import repr_type
from telegrinder.node.composer import NodeSession, compose_nodes
from telegrinder.types.objects import Update

if typing.TYPE_CHECKING:
    from telegrinder.node.base import Node

Ts = typing.TypeVarTuple("Ts", default=typing.Unpack[tuple[type["Node"], ...]])


class NodeAdapter(typing.Generic[*Ts], ABCAdapter[Update, Event[tuple[*Ts]]]):
    def __init__(self, *nodes: *Ts) -> None:
        self.nodes = nodes

    def __repr__(self) -> str:
        return "<{}: adapt Update -> ({})>".format(
            self.__class__.__name__,
            ", ".join(repr_type(node) for node in self.nodes),
        )

    async def adapt(
        self,
        api: API,
        update: Update,
        context: Context,
    ) -> Result[Event[tuple[*Ts]], AdapterError]:
        result = await compose_nodes(
            nodes={str(i): typing.cast(type["Node"], node) for i, node in enumerate(self.nodes)},
            ctx=context,
            data={Update: update, API: api},
        )

        match result:
            case Ok(collection):
                sessions: list[NodeSession] = list(collection.sessions.values())
                return Ok(Event(tuple(sessions)))  # type: ignore
            case Error(err):
                return Error(AdapterError(f"Couldn't compose nodes, error: {err}."))


__all__ = ("NodeAdapter",)



import abc
import inspect
import typing
from types import AsyncGeneratorType

from telegrinder.node.scope import NodeScope
from telegrinder.tools.magic import (
    cache_magic_value,
    get_annotations,
)

ComposeResult: typing.TypeAlias = typing.Awaitable[typing.Any] | typing.AsyncGenerator[typing.Any, None]


def is_node(maybe_node: typing.Any) -> typing.TypeGuard[type["Node"]]:
    maybe_node = maybe_node if isinstance(maybe_node, type) else typing.get_origin(maybe_node)
    return (
        isinstance(maybe_node, type)
        and issubclass(maybe_node, Node)
        or isinstance(maybe_node, Node)
        or hasattr(maybe_node, "as_node")
    )


@cache_magic_value("__nodes__")
def get_nodes(function: typing.Callable[..., typing.Any]) -> dict[str, type["Node"]]:
    return {k: v for k, v in get_annotations(function).items() if is_node(v)}


@cache_magic_value("__is_generator__")
def is_generator(function: typing.Callable[..., typing.Any]) -> typing.TypeGuard[AsyncGeneratorType[typing.Any, None]]:
    return inspect.isasyncgenfunction(function)


def get_node_calc_lst(node: type["Node"]) -> list[type["Node"]]:
    """Returns flattened list of node types in ordering required to calculate given node.
    Provides caching for passed node type"""

    if calc_lst := getattr(node, "__nodes_calc_lst__", None):
        return calc_lst
    nodes_lst: list[type["Node"]] = []
    annotations = list(node.as_node().get_subnodes().values())
    for node_type in annotations:
        nodes_lst.extend(get_node_calc_lst(node_type))
    calc_lst = [*nodes_lst, node]
    setattr(node, "__nodes_calc_lst__", calc_lst)
    return calc_lst


class ComposeError(BaseException):
    pass


class Node(abc.ABC):
    node: str = "node"
    scope: NodeScope = NodeScope.PER_EVENT

    @classmethod
    @abc.abstractmethod
    def compose(cls, *args, **kwargs) -> ComposeResult:
        pass

    @classmethod
    def compose_error(cls, error: str | None = None) -> typing.NoReturn:
        raise ComposeError(error)

    @classmethod
    def get_subnodes(cls) -> dict[str, type["Node"]]:
        return get_nodes(cls.compose)

    @classmethod
    def as_node(cls) -> type[typing.Self]:
        return cls

    @classmethod
    def is_generator(cls) -> bool:
        return is_generator(cls.compose)


class DataNode(Node, abc.ABC):
    node = "data"

    @typing.dataclass_transform()
    @classmethod
    @abc.abstractmethod
    async def compose(cls, *args, **kwargs) -> ComposeResult:
        pass


class ScalarNodeProto(Node, abc.ABC):
    @classmethod
    @abc.abstractmethod
    async def compose(cls, *args, **kwargs) -> ComposeResult:
        pass


SCALAR_NODE = type("ScalarNode", (), {"node": "scalar"})


if typing.TYPE_CHECKING:

    class ScalarNode(ScalarNodeProto, abc.ABC):
        pass

else:
    def __init_subclass__(cls, *args, **kwargs):  # noqa: N807
        if any(issubclass(base, ScalarNode) for base in cls.__bases__ if base is not ScalarNode):
            raise RuntimeError("Scalar nodes do not support inheritance.")

    def create_node(cls, bases, dct):
        dct.update(cls.__dict__)
        return type(cls.__name__, bases, dct)

    def create_class(name, bases, dct):
        return type(
            "Scalar",
            (SCALAR_NODE,),
            {
                "as_node": classmethod(lambda cls: create_node(cls, bases, dct)),
                "scope": Node.scope,
                "__init_subclass__": __init_subclass__,
            },
        )

    class ScalarNode(ScalarNodeProto, abc.ABC, metaclass=create_class):
        pass


__all__ = (
    "ComposeError",
    "DataNode",
    "Node",
    "SCALAR_NODE",
    "ScalarNode",
    "ScalarNodeProto",
    "get_nodes",
    "is_node",
)



import typing

from telegrinder.msgspec_utils import decoder, encoder


def loads(s: str | bytes) -> typing.Any:
    return decoder.decode(s)


def dumps(o: typing.Any) -> str:
    return encoder.encode(o)


__all__ = ("dumps", "loads")



import inspect
import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.types.objects import Update

from .abc import ABCAdapter, ABCRule, AdaptTo, RawUpdateAdapter


class FuncRule(ABCRule, typing.Generic[AdaptTo]):
    def __init__(
        self,
        func: typing.Callable[[AdaptTo, Context], typing.Awaitable[bool] | bool],
        adapter: ABCAdapter[Update, AdaptTo] | None = None,
    ) -> None:
        self.func = func
        self.adapter = adapter or RawUpdateAdapter()  # type: ignore

    async def check(self, event: AdaptTo, ctx: Context) -> bool:
        result = self.func(event, ctx)
        if inspect.isawaitable(result):
            return await result
        return result  # type: ignore


__all__ = ("FuncRule",)



import typing
from abc import ABC

from telegrinder.bot.dispatch.context import Context
from telegrinder.model import Model

Event = typing.TypeVar("Event", bound=Model)


class ABCMiddleware(ABC, typing.Generic[Event]):
    async def pre(self, event: Event, ctx: Context) -> bool: ...

    async def post(self, event: Event, responses: list[typing.Any], ctx: Context) -> None: ...


__all__ = ("ABCMiddleware",)



import dataclasses
import typing

import msgspec

from telegrinder.api import API
from telegrinder.bot.cute_types import BaseCute
from telegrinder.bot.dispatch.context import Context
from telegrinder.msgspec_utils import DataclassInstance, decoder
from telegrinder.node.base import ComposeError, Node
from telegrinder.node.update import UpdateNode

if typing.TYPE_CHECKING:
    Dataclass = typing.TypeVar("Dataclass", bound="DataclassType")

    DataclassType: typing.TypeAlias = DataclassInstance | msgspec.Struct | dict[str, typing.Any]

EVENT_NODE_KEY = "_event_node"


class _EventNode(Node):
    dataclass: type["DataclassType"]

    def __new__(cls, dataclass: type["DataclassType"], /) -> type[typing.Self]:
        namespace = dict(**cls.__dict__)
        namespace.pop("__new__", None)
        new_cls = type("EventNode", (cls,), {"dataclass": dataclass, **namespace})
        return new_cls  # type: ignore

    def __class_getitem__(cls, dataclass: type["DataclassType"], /) -> typing.Self:
        return cls(dataclass)

    @classmethod
    async def compose(cls, raw_update: UpdateNode, ctx: Context, api: API) -> "DataclassType":
        dataclass_type = typing.get_origin(cls.dataclass) or cls.dataclass

        try:
            if issubclass(dataclass_type, dict):
                dataclass = cls.dataclass(**raw_update.incoming_update.to_full_dict())

            elif issubclass(dataclass_type, BaseCute):
                if isinstance(raw_update.incoming_update, dataclass_type):
                    dataclass = raw_update.incoming_update
                else:
                    dataclass = dataclass_type.from_update(raw_update.incoming_update, bound_api=api)

            elif issubclass(dataclass_type, msgspec.Struct) or dataclasses.is_dataclass(dataclass_type):
                # FIXME: must be used with rename_field
                dataclass = decoder.convert(
                    raw_update.incoming_update.to_full_dict(),
                    type=cls.dataclass,
                    str_keys=True,
                )

            else:
                dataclass = cls.dataclass(**raw_update.incoming_update.to_dict())

            ctx[EVENT_NODE_KEY] = cls
            return dataclass
        except Exception as exc:
            raise ComposeError(f"Cannot parse update into {cls.dataclass.__name__!r}, error: {exc}")


if typing.TYPE_CHECKING:
    EventNode: typing.TypeAlias = typing.Annotated["Dataclass", ...]

else:

    class EventNode(_EventNode):
        pass


__all__ = ("EventNode",)



import typing
from abc import ABC, abstractmethod

from telegrinder.api import API
from telegrinder.bot.dispatch.context import Context
from telegrinder.model import Model
from telegrinder.types.objects import Update

T = typing.TypeVar("T", bound=Model)


class ABCHandler(ABC, typing.Generic[T]):
    is_blocking: bool

    @abstractmethod
    async def check(self, api: API, event: Update, ctx: Context | None = None) -> bool:
        pass

    @abstractmethod
    async def run(self, api: API, event: T, ctx: Context) -> typing.Any:
        pass


__all__ = ("ABCHandler",)



import dataclasses

import typing_extensions as typing

from telegrinder.bot.dispatch.view import (
    callback_query,
    chat_join_request,
    chat_member,
    inline_query,
    message,
    raw,
)
from telegrinder.bot.dispatch.view.abc import ABCEventRawView, ABCView
from telegrinder.types.enums import UpdateType

CallbackQueryView = typing.TypeVar(
    "CallbackQueryView", bound=ABCView, default=callback_query.CallbackQueryView
)
ChatJoinRequestView = typing.TypeVar(
    "ChatJoinRequestView", bound=ABCView, default=chat_join_request.ChatJoinRequestView
)
ChatMemberView = typing.TypeVar("ChatMemberView", bound=ABCView, default=chat_member.ChatMemberView)
InlineQueryView = typing.TypeVar("InlineQueryView", bound=ABCView, default=inline_query.InlineQueryView)
MessageView = typing.TypeVar("MessageView", bound=ABCView, default=message.MessageView)
RawEventView = typing.TypeVar("RawEventView", bound=ABCEventRawView, default=raw.RawEventView)


@dataclasses.dataclass(kw_only=True)
class ViewBox(
    typing.Generic[
        CallbackQueryView,
        ChatJoinRequestView,
        ChatMemberView,
        InlineQueryView,
        MessageView,
        RawEventView,
    ],
):
    callback_query_view: dataclasses.InitVar[CallbackQueryView | None] = None
    chat_join_request_view: dataclasses.InitVar[ChatJoinRequestView | None] = None
    chat_member_view: dataclasses.InitVar[ChatMemberView | None] = None
    my_chat_member_view: dataclasses.InitVar[ChatMemberView | None] = None
    inline_query_view: dataclasses.InitVar[InlineQueryView | None] = None
    message_view: dataclasses.InitVar[MessageView | None] = None
    business_message_view: dataclasses.InitVar[MessageView | None] = None
    channel_post_view: dataclasses.InitVar[MessageView | None] = None
    edited_message_view: dataclasses.InitVar[MessageView | None] = None
    edited_business_message_view: dataclasses.InitVar[MessageView | None] = None
    edited_channel_post_view: dataclasses.InitVar[MessageView | None] = None
    any_message_view: dataclasses.InitVar[MessageView | None] = None
    chat_member_updated_view: dataclasses.InitVar[ChatMemberView | None] = None
    raw_event_view: dataclasses.InitVar[RawEventView | None] = None

    def __post_init__(
        self,
        callback_query_view: CallbackQueryView | None = None,
        chat_join_request_view: ChatJoinRequestView | None = None,
        chat_member_view: ChatMemberView | None = None,
        my_chat_member_view: ChatMemberView | None = None,
        inline_query_view: InlineQueryView | None = None,
        message_view: MessageView | None = None,
        business_message_view: MessageView | None = None,
        channel_post_view: MessageView | None = None,
        edited_message_view: MessageView | None = None,
        edited_business_message_view: MessageView | None = None,
        edited_channel_post_view: MessageView | None = None,
        any_message_view: MessageView | None = None,
        chat_member_updated_view: ChatMemberView | None = None,
        raw_event_view: RawEventView | None = None,
    ) -> None:
        self.callback_query = typing.cast(
            CallbackQueryView,
            callback_query_view or callback_query.CallbackQueryView(),
        )
        self.chat_join_request = typing.cast(
            ChatJoinRequestView,
            chat_join_request_view or chat_join_request.ChatJoinRequestView(),
        )
        self.chat_member = typing.cast(
            ChatMemberView,
            chat_member_view or chat_member.ChatMemberView(update_type=UpdateType.CHAT_MEMBER),
        )
        self.my_chat_member = typing.cast(
            ChatMemberView,
            my_chat_member_view or chat_member.ChatMemberView(update_type=UpdateType.MY_CHAT_MEMBER),
        )
        self.inline_query = typing.cast(
            InlineQueryView,
            inline_query_view or inline_query.InlineQueryView(),
        )
        self.message = typing.cast(
            MessageView,
            message_view or message.MessageView(update_type=UpdateType.MESSAGE),
        )
        self.business_message = typing.cast(
            MessageView,
            business_message_view or message.MessageView(update_type=UpdateType.BUSINESS_MESSAGE),
        )
        self.channel_post = typing.cast(
            MessageView,
            channel_post_view or message.MessageView(update_type=UpdateType.CHANNEL_POST),
        )
        self.edited_message = typing.cast(
            MessageView,
            edited_message_view or message.MessageView(update_type=UpdateType.EDITED_MESSAGE),
        )
        self.edited_business_message = typing.cast(
            MessageView,
            edited_business_message_view
            or message.MessageView(update_type=UpdateType.EDITED_BUSINESS_MESSAGE),
        )
        self.edited_channel_post = typing.cast(
            MessageView,
            edited_channel_post_view or message.MessageView(update_type=UpdateType.EDITED_CHANNEL_POST),
        )
        self.any_message = typing.cast(MessageView, any_message_view or message.MessageView())
        self.chat_member_updated = typing.cast(
            ChatMemberView,
            chat_member_updated_view or chat_member.ChatMemberView(),
        )
        self.raw_event = typing.cast(RawEventView, raw_event_view or raw.RawEventView())

    def get_views(self) -> dict[str, ABCView]:
        """Get all views."""

        return {name: view for name, view in self.__dict__.items() if isinstance(view, ABCView)}


__all__ = ("ViewBox",)



import typing_extensions as typing

from telegrinder.api.api import API
from telegrinder.bot.dispatch.abc import ABCDispatch
from telegrinder.bot.dispatch.dispatch import Dispatch
from telegrinder.bot.polling.abc import ABCPolling
from telegrinder.bot.polling.polling import Polling
from telegrinder.modules import logger
from telegrinder.tools.loop_wrapper import ABCLoopWrapper
from telegrinder.tools.loop_wrapper.loop_wrapper import LoopWrapper

DispatchT = typing.TypeVar("DispatchT", bound=ABCDispatch, default=Dispatch)
PollingT = typing.TypeVar("PollingT", bound=ABCPolling, default=Polling)
LoopWrapperT = typing.TypeVar("LoopWrapperT", bound=ABCLoopWrapper, default=LoopWrapper)


class Telegrinder(typing.Generic[DispatchT, PollingT, LoopWrapperT]):
    def __init__(
        self,
        api: API,
        *,
        polling: PollingT | None = None,
        dispatch: DispatchT | None = None,
        loop_wrapper: LoopWrapperT | None = None,
    ) -> None:
        self.api = api
        self.dispatch = typing.cast(DispatchT, dispatch or Dispatch())
        self.polling = typing.cast(PollingT, polling or Polling(api))
        self.loop_wrapper = typing.cast(LoopWrapperT, loop_wrapper or LoopWrapper())

    def __repr__(self) -> str:
        return "<{}: api={!r}, dispatch={!r}, polling={!r}, loop_wrapper={!r}>".format(
            self.__class__.__name__,
            self.api,
            self.dispatch,
            self.polling,
            self.loop_wrapper,
        )

    @property
    def on(self) -> DispatchT:
        return self.dispatch

    async def reset_webhook(self) -> None:
        if not (await self.api.get_webhook_info()).unwrap().url:
            return
        await self.api.delete_webhook()

    async def run_polling(
        self,
        *,
        offset: int = 0,
        skip_updates: bool = False,
    ) -> None:
        if skip_updates:
            logger.debug("Dropping pending updates")
            await self.reset_webhook()
            await self.api.delete_webhook(drop_pending_updates=True)
        self.polling.offset = offset

        async for updates in self.polling.listen():
            for update in updates:
                logger.debug(
                    "Received update (update_id={}, update_type={!r})",
                    update.update_id,
                    update.update_type.name,
                )
                self.loop_wrapper.add_task(self.dispatch.feed(update, self.api))

    def run_forever(self, *, offset: int = 0, skip_updates: bool = False) -> None:
        logger.debug("Running blocking polling (id={})", self.api.id)
        self.loop_wrapper.add_task(self.run_polling(offset=offset, skip_updates=skip_updates))
        self.loop_wrapper.run_event_loop()


__all__ = ("Telegrinder",)



from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.node.base import ComposeError, ScalarNode
from telegrinder.node.update import UpdateNode


class CallbackQueryNode(ScalarNode, CallbackQueryCute):
    @classmethod
    async def compose(cls, update: UpdateNode) -> CallbackQueryCute:
        if not update.callback_query:
            raise ComposeError("Update is not a callback_query.")
        return update.callback_query.unwrap()


__all__ = ("CallbackQueryNode",)



from abc import ABC, abstractmethod


class KeyboardSetError(LookupError):
    pass


class KeyboardSetBase(ABC):
    @classmethod
    @abstractmethod
    def load(cls) -> None:
        pass


__all__ = ("KeyboardSetBase", "KeyboardSetError")



from telegrinder.bot.cute_types.chat_join_request import ChatJoinRequestCute
from telegrinder.bot.dispatch.view.base import BaseStateView


class ChatJoinRequestView(BaseStateView[ChatJoinRequestCute]):
    def __init__(self) -> None:
        self.auto_rules = []
        self.handlers = []
        self.middlewares = []
        self.return_manager = None

    def get_state_key(self, event: ChatJoinRequestCute) -> int | None:
        return event.chat_id


__all__ = ("ChatJoinRequestView",)



from telegrinder.node.text import TextInteger

from .abc import ABCRule
from .node import NodeRule


class IsInteger(NodeRule):
    def __init__(self) -> None:
        super().__init__(TextInteger)


class IntegerInRange(ABCRule):
    def __init__(self, rng: range) -> None:
        self.rng = rng

    async def check(self, integer: TextInteger) -> bool:
        return integer in self.rng


__all__ = ("IntegerInRange", "IsInteger")



from telegrinder.bot.cute_types.base import BaseCute
from telegrinder.bot.cute_types.callback_query import CallbackQueryCute
from telegrinder.bot.cute_types.chat_join_request import ChatJoinRequestCute
from telegrinder.bot.cute_types.chat_member_updated import ChatMemberUpdatedCute
from telegrinder.bot.cute_types.inline_query import InlineQueryCute
from telegrinder.bot.cute_types.message import MessageCute
from telegrinder.bot.cute_types.update import UpdateCute

__all__ = (
    "BaseCute",
    "CallbackQueryCute",
    "ChatJoinRequestCute",
    "ChatMemberUpdatedCute",
    "InlineQueryCute",
    "MessageCute",
    "UpdateCute",
)



import typing

from telegrinder.api.api import API
from telegrinder.bot.cute_types.update import UpdateCute
from telegrinder.bot.dispatch.handler.func import FuncHandler
from telegrinder.bot.dispatch.process import process_inner
from telegrinder.bot.dispatch.view.abc import ABCEventRawView
from telegrinder.bot.dispatch.view.base import BaseView, ErrorHandlerT
from telegrinder.bot.rules.abc import ABCRule
from telegrinder.tools.error_handler.error_handler import ABCErrorHandler, ErrorHandler
from telegrinder.types.enums import UpdateType
from telegrinder.types.objects import Update

T = typing.TypeVar("T")

FuncType: typing.TypeAlias = typing.Callable[
    typing.Concatenate[T, ...],
    typing.Coroutine[typing.Any, typing.Any, typing.Any],
]


class RawEventView(ABCEventRawView[UpdateCute], BaseView[UpdateCute]):
    def __init__(self) -> None:
        self.auto_rules = []
        self.handlers = []
        self.middlewares = []
        self.return_manager = None

    @typing.overload
    def __call__(
        self,
        update_type: UpdateType,
        *rules: ABCRule,
    ) -> typing.Callable[
        [FuncType[UpdateCute]],
        FuncHandler[UpdateCute, FuncType[UpdateCute], ErrorHandler[UpdateCute]],
    ]: ...

    @typing.overload
    def __call__(
        self,
        update_type: UpdateType,
        *rules: ABCRule,
        dataclass: type[T],
    ) -> typing.Callable[[FuncType[T]], FuncHandler[UpdateCute, FuncType[T], ErrorHandler[T]]]: ...

    @typing.overload
    def __call__(
        self,
        update_type: UpdateType,
        *rules: ABCRule,
        error_handler: ErrorHandlerT,
    ) -> typing.Callable[
        [FuncType[UpdateCute]],
        FuncHandler[UpdateCute, FuncType[UpdateCute], ErrorHandlerT],
    ]: ...

    @typing.overload
    def __call__(
        self,
        update_type: UpdateType,
        *rules: ABCRule,
        dataclass: type[T],
        error_handler: ErrorHandlerT,
        is_blocking: bool = True,
    ) -> typing.Callable[[FuncType[T]], FuncHandler[UpdateCute, FuncType[T], ErrorHandlerT]]: ...

    @typing.overload
    def __call__(
        self,
        update_type: UpdateType,
        *rules: ABCRule,
        dataclass: typing.Literal[None] = None,
        error_handler: typing.Literal[None] = None,
        is_blocking: bool = True,
    ) -> typing.Callable[
        [FuncType[UpdateCute]],
        FuncHandler[UpdateCute, FuncType[UpdateCute], ErrorHandler[UpdateCute]],
    ]: ...

    def __call__(  # type: ignore
        self,
        update_type: UpdateType,
        *rules: ABCRule,
        dataclass: type[typing.Any] | None = None,
        error_handler: ABCErrorHandler | None = None,
        is_blocking: bool = True,
    ):
        def wrapper(func: FuncType[typing.Any]):
            func_handler = FuncHandler(
                func,
                rules=[*self.auto_rules, *rules],
                is_blocking=is_blocking,
                dataclass=dataclass,
                error_handler=error_handler or ErrorHandler(),
                update_type=update_type,
            )
            self.handlers.append(func_handler)
            return func_handler

        return wrapper

    async def check(self, event: Update) -> bool:
        return bool(self.handlers) or bool(self.middlewares)

    async def process(self, event: Update, api: API) -> bool:
        return await process_inner(
            api,
            UpdateCute.from_update(event, bound_api=api),
            event,
            self.middlewares,
            self.handlers,
            self.return_manager,
        )



from fntypes import Option

from telegrinder.bot.dispatch.view import BaseStateView
from telegrinder.tools.functional import from_optional

from .hasher import ECHO, Event, Hasher


class StateViewHasher(Hasher[Event, int]):
    view: BaseStateView

    def __init__(self, view: type[BaseStateView[Event]]):
        super().__init__(view, get_hash_from_data=ECHO)

    def get_data_from_event(self, event: Event) -> Option[int]:
        return from_optional(self.view.get_state_key(event))


__all__ = ("StateViewHasher",)



"""Telegrinder

Modern visionary telegram bot framework.

* Type hinted
* Customizable and extensible
* Ready to use scenarios and rules
* Fast models built on [msgspec](https://github.com/jcrist/msgspec)
* Both low-level and high-level API
* Support [optional dependecies](https://github.com/timoniq/telegrinder/blob/dev/docs/guide/optional_dependencies.md)

Basic example:

```python
from telegrinder import API, Message, Telegrinder, Token
from telegrinder.modules import logger
from telegrinder.rules import Text

api = API(token=Token("123:token"))
bot = Telegrinder(api)
logger.set_level("INFO")


@bot.on.message(Text("/start"))
async def start(message: Message):
    me = (await api.get_me()).unwrap()
    await message.answer(f"Hello, {message.from_user.full_name}! I'm {me.full_name}.")


bot.run_forever()
```
"""

import typing

from .api import API, APIError, APIResponse, Token
from .bot import (
    CALLBACK_QUERY_FOR_MESSAGE,
    CALLBACK_QUERY_FROM_CHAT,
    CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE,
    MESSAGE_FROM_USER,
    MESSAGE_FROM_USER_IN_CHAT,
    MESSAGE_IN_CHAT,
    ABCDispatch,
    ABCHandler,
    ABCMiddleware,
    ABCPolling,
    ABCReturnManager,
    ABCRule,
    ABCScenario,
    ABCStateView,
    ABCView,
    AudioReplyHandler,
    BaseCute,
    BaseReturnManager,
    BaseStateView,
    BaseView,
    CallbackQueryCute,
    CallbackQueryReturnManager,
    CallbackQueryRule,
    CallbackQueryView,
    ChatJoinRequestCute,
    ChatJoinRequestRule,
    ChatJoinRequestView,
    ChatMemberUpdatedCute,
    ChatMemberView,
    Checkbox,
    Choice,
    Context,
    Dispatch,
    DocumentReplyHandler,
    FuncHandler,
    Hasher,
    InlineQueryCute,
    InlineQueryReturnManager,
    InlineQueryRule,
    MediaGroupReplyHandler,
    MessageCute,
    MessageReplyHandler,
    MessageReturnManager,
    MessageRule,
    MessageView,
    PhotoReplyHandler,
    Polling,
    RawEventView,
    ShortState,
    StateViewHasher,
    StickerReplyHandler,
    Telegrinder,
    UpdateCute,
    VideoReplyHandler,
    ViewBox,
    WaiterMachine,
    register_manager,
)
from .bot.rules import StateMeta
from .client import ABCClient, AiohttpClient
from .model import Model
from .modules import logger
from .tools import (
    ABCErrorHandler,
    ABCGlobalContext,
    ABCLoopWrapper,
    ABCStateStorage,
    ABCTranslator,
    ABCTranslatorMiddleware,
    AnyMarkup,
    Button,
    CtxVar,
    DelayedTask,
    ErrorHandler,
    FormatString,
    GlobalContext,
    HTMLFormatter,
    I18nEnum,
    InlineButton,
    InlineKeyboard,
    Keyboard,
    KeyboardSetBase,
    KeyboardSetYAML,
    Lifespan,
    LoopWrapper,
    MemoryStateStorage,
    ParseMode,
    RowButtons,
    SimpleI18n,
    SimpleTranslator,
    StateData,
    ctx_var,
    magic_bundle,
)

Update: typing.TypeAlias = UpdateCute
Message: typing.TypeAlias = MessageCute
ChatJoinRequest: typing.TypeAlias = ChatJoinRequestCute
ChatMemberUpdated: typing.TypeAlias = ChatMemberUpdatedCute
CallbackQuery: typing.TypeAlias = CallbackQueryCute
InlineQuery: typing.TypeAlias = InlineQueryCute
Bot: typing.TypeAlias = Telegrinder


__all__ = (
    "ABCClient",
    "ABCDispatch",
    "ABCErrorHandler",
    "ABCGlobalContext",
    "ABCHandler",
    "ABCLoopWrapper",
    "ABCMiddleware",
    "ABCPolling",
    "ABCReturnManager",
    "ABCRule",
    "ABCScenario",
    "ABCStateStorage",
    "ABCStateStorage",
    "ABCStateView",
    "ABCTranslator",
    "ABCTranslatorMiddleware",
    "ABCView",
    "API",
    "APIError",
    "APIResponse",
    "AiohttpClient",
    "AnyMarkup",
    "AudioReplyHandler",
    "BaseCute",
    "BaseReturnManager",
    "BaseStateView",
    "BaseView",
    "Bot",
    "Button",
    "CALLBACK_QUERY_FOR_MESSAGE",
    "CALLBACK_QUERY_FROM_CHAT",
    "CALLBACK_QUERY_IN_CHAT_FOR_MESSAGE",
    "CallbackQuery",
    "CallbackQueryCute",
    "CallbackQueryReturnManager",
    "CallbackQueryRule",
    "CallbackQueryView",
    "ChatJoinRequest",
    "ChatJoinRequestCute",
    "ChatJoinRequestRule",
    "ChatJoinRequestView",
    "ChatMemberUpdated",
    "ChatMemberUpdatedCute",
    "ChatMemberView",
    "DocumentReplyHandler",
    "Checkbox",
    "Choice",
    "Context",
    "CtxVar",
    "DelayedTask",
    "Dispatch",
    "ErrorHandler",
    "FormatString",
    "FuncHandler",
    "GlobalContext",
    "HTMLFormatter",
    "Hasher",
    "I18nEnum",
    "InlineButton",
    "InlineKeyboard",
    "InlineQuery",
    "InlineQueryCute",
    "InlineQueryReturnManager",
    "InlineQueryRule",
    "Keyboard",
    "KeyboardSetBase",
    "KeyboardSetYAML",
    "Lifespan",
    "LoopWrapper",
    "MESSAGE_FROM_USER",
    "MESSAGE_FROM_USER_IN_CHAT",
    "MESSAGE_IN_CHAT",
    "MediaGroupReplyHandler",
    "MemoryStateStorage",
    "MemoryStateStorage",
    "Message",
    "MessageCute",
    "MessageReplyHandler",
    "MessageReplyHandler",
    "MessageReturnManager",
    "MessageRule",
    "MessageView",
    "Model",
    "ParseMode",
    "PhotoReplyHandler",
    "Polling",
    "RawEventView",
    "RowButtons",
    "ShortState",
    "SimpleI18n",
    "SimpleTranslator",
    "StateData",
    "StateData",
    "StateMeta",
    "StateMeta",
    "StateViewHasher",
    "StickerReplyHandler",
    "Telegrinder",
    "Token",
    "Update",
    "UpdateCute",
    "VideoReplyHandler",
    "ViewBox",
    "WaiterMachine",
    "ctx_var",
    "logger",
    "magic_bundle",
    "register_manager",
)



import typing

from telegrinder.node.base import Node


class ContainerNode(Node):
    linked_nodes: typing.ClassVar[list[type[Node]]]

    @classmethod
    async def compose(cls, **kw) -> tuple[Node, ...]:
        return tuple(t[1] for t in sorted(kw.items(), key=lambda t: t[0]))

    @classmethod
    def get_subnodes(cls) -> dict[str, type[Node]]:
        subnodes = getattr(cls, "subnodes", None)
        if subnodes is None:
            subnodes_dct = {f"node_{i}": node_t for i, node_t in enumerate(cls.linked_nodes)}
            setattr(cls, "subnodes", subnodes_dct)
            return subnodes_dct
        return subnodes

    @classmethod
    def link_nodes(cls, linked_nodes: list[type[Node]]) -> type["ContainerNode"]:
        return type("_ContainerNode", (cls,), {"linked_nodes": linked_nodes})


__all__ = ("ContainerNode",)



from .abc import ABCErrorHandler
from .error import CatcherError
from .error_handler import Catcher, ErrorHandler

__all__ = (
    "ABCErrorHandler",
    "Catcher",
    "CatcherError",
    "ErrorHandler",
)



import typing

from telegrinder.bot.dispatch.context import Context
from telegrinder.types.enums import MessageEntityType
from telegrinder.types.objects import MessageEntity

from .message import Message, MessageRule

Entity: typing.TypeAlias = str | MessageEntityType


class HasEntities(MessageRule):
    async def check(self, message: Message) -> bool:
        return bool(message.entities)


class MessageEntities(MessageRule, requires=[HasEntities()]):
    def __init__(self, entities: Entity | list[Entity], /) -> None:
        self.entities = [entities] if not isinstance(entities, list) else entities

    async def check(self, message: Message, ctx: Context) -> bool:
        message_entities: list[MessageEntity] = []
        for entity in message.entities.unwrap():
            for entity_type in self.entities:
                if entity_type == entity.type:
                    message_entities.append(entity)

        if not message_entities:
            return False

        ctx.message_entities = message_entities
        return True


__all__ = ("HasEntities", "MessageEntities")
